@@ -1,4 +1,4 @@
-# $Id: Base.pm 97 2005-11-15 15:22:35Z daisuke $
+# $Id: Base.pm 107 2006-05-17 09:01:19Z daisuke $
 #
 # Copyright (c) 2005 Daisuke Maki <dmaki@cpan.org>
 # All rights reserved.
@@ -40,9 +40,27 @@ sub initialize
         $self->_load_config($config);
     }
 
-    # Set values that must be set
+    # Set values provided by %args. This overrides the config file
+    $self->{ALIAS}                = $args{Alias} if exists $args{Alias};
+    $self->{DNS_CACHE_ARGS}       = $args{DnsCacheArgs}
+        if exists $args{DnsCacheArgs};
+    $self->{DNS_CACHE_CLASS}      = $args{DnsCacheClass}
+        if exists $args{DnsCacheClass};
+    $self->{DNS_CACHE_ARGS_DEREF} = $args{DnsCacheArgsDeref}
+        if $args{DnsCacheArgsDeref};
+    $self->{DNS_COMP_ALIAS}       = $args{DnsCompAlias} if $args{DnsCompAlias};
+    $self->{DNS_COMP_CLASS}       = $args{DnsCompClass} if $args{DnsCompClass};
+    $self->{HANDLER_ALIAS}        = $args{HandlerAlias} if $args{HandlerAlias};
+    $self->{HTTP_COMP_CLASS}      = $args{HttpCompClass} if $args{HttpCompClass};
+    $self->{HTTP_COMP_ARGS}       = $args{HttpCompArgs} if $args{HttpCompArgs};
+    $self->{MAX_HTTP_COMP}        = $args{MaxHttpComp} if $args{MaxHttpComp};
+    $self->{ENABLE_DNS_CACHE}     = exists $args{EnableDnsCache} ?
+        $args{EnableDnsCache} : 1;
+
+    # Set defaults.
     $self->{ALIAS}             ||= 'broker';
     $self->{CONFIG_FILE}       ||= undef;
+    $self->{DNS_CACHE}         ||= undef;
     $self->{DNS_CACHE_CLASS}   ||= DEFAULT_DNS_CACHE_CLASS;
     $self->{DNS_CACHE_ARGS}    ||= DEFAULT_DNS_CACHE_ARGS;
     if (!exists $self->{DNS_CACHE_ARGS_DEREF}) {
@@ -60,6 +78,7 @@ sub initialize
     $self->{SHUTDOWN}            = 0;  # Non-configurable
     $self->{STRICT_JOB_TYPE}   ||= 'Xango::Job';
     $self->{UNDEF_ON_FINALIZE} ||= 0;
+    $self->{STATISTICS}        ||= {};
 
     my $k_munge;
     foreach my $k (keys %args) {
@@ -67,6 +86,15 @@ sub initialize
         $k_munge =~ s/([a-z])([A-Z])/$1_$2/g;
         $self->{uc ($k_munge)} = $args{$k};
     }
+
+    $self->{AUTO_THROTTLE} ||= 0;
+    if ($self->{AUTO_THROTTLE}) {
+        require Data::Average::BoundedExpires;
+        $self->{THROTTLE_DATA} ||= Data::Average::BoundedExpires->new(
+            max => 100,
+            expires_in => 60
+        );
+    }
 }
 
 sub _load_config
@@ -83,6 +111,7 @@ sub _load_config
         $c = uc $c;
         $self->{$c} = $config->param($p);
     };
+    $set->('AutoThrottle');
     $set->('DnsCacheClass');
     $set->('DnsCacheArgs');
     $set->('DnsCacheArgsDeref');
@@ -119,6 +148,7 @@ sub states
                 'spawn_dns_comp',
                 'spawn_http_comp',
                 'unregister_http_request',
+                'shutdown_broker',
             ]
         ]
     );
@@ -172,8 +202,12 @@ sub _start
     # Spawn a DNS component to resolve hostnames
     $kernel->call($session, 'spawn_dns_comp');
 
-    my $cache = $kernel->call($session, 'create_dns_cache');
-    $obj->attr(dns_cache => $cache);
+    # create DNS cache, if specified.
+    if ($obj->enable_dns_cache) {
+        my $cache = $kernel->call($session, 'create_dns_cache');
+        $cache->clear();
+        $obj->attr(dns_cache => $cache);
+    }
 
     $kernel->call(
         $session,
@@ -190,6 +224,12 @@ sub _stop
     $kernel->alias_remove($obj->alias);
 }
 
+sub shutdown_broker
+{
+    my($obj) = @_[OBJECT];
+    $obj->shutdown(1);
+}
+
 sub spawn_dns_comp
 {
     my($kernel, $object) = @_[KERNEL, OBJECT];
@@ -288,7 +328,7 @@ sub create_http_request
     $req->header(Host => $job->notes('host_name'));
 
     # Give the handler a chance to munge with the request
-    $kernel->call($session, 'prep_request', $job, $req);
+    $kernel->call($obj->handler_alias, 'prep_request', $job, $req);
 
     return $req;
 }
@@ -321,20 +361,67 @@ sub unregister_http_request
     my $fetchers = $obj->fetchers();
     my $data     = $fetchers->{$fetcher_id};
 
-    $data->{last_request_time} = time();
-    delete $data->{jobs}->{$job->id};
+    if ($data) {
+        $data->{last_request_time} = time();
+        delete $data->{jobs}->{$job->id};
+    }
 }
 
 sub dispatch_to_lightest_load
 {
     my($kernel, $session, $obj, $job) = @_[KERNEL, SESSION, OBJECT, ARG0];
 
+    my %fetchers = %{ $obj->fetchers };
+
+    # throttle if necessary
+    if ($obj->auto_throttle && $obj->throttle_data->length > 10) {
+        my $avg = $obj->throttle_data->avg;
+
+        my $max_http_comp = $obj->max_http_comp;
+        if ($avg < 1) {
+            my $target =
+                $avg < 0.5 ? int($max_http_comp * 0.1) :
+                $avg < 0.8 ? int($max_http_comp * 0.2) :
+                $avg < 0.9 ? int($max_http_comp * 0.5) :
+                $max_http_comp;
+
+            # count fetchers that aren't marked as to be deleted
+            my @active = grep { ! $_->{shutdown} } keys %fetchers;
+            if (@active > $max_http_comp) {
+                # randomly select a few fetchers and mark them as to be
+                # shutdown soon
+                for my $i (1..$max_http_comp - scalar @active) {
+                    $fetchers{ $active[rand(@active)]->{id} }->{shutdown}++;
+                }
+            }
+        } else {
+            if ($max_http_comp > keys %fetchers) {
+                # spawn more, if we've throttled it down
+                # only spawn up to 1/10 of the max_http_comp size
+
+                my $missing  = $max_http_comp - keys %fetchers;
+                my $to_spawn =
+                    $max_http_comp <= 10 ? 1 :
+                    $missing < int($max_http_comp / 10) ? $missing :
+                    int($max_http_comp / 10);
+
+                # synchronous, cause we might accidentally spawn more
+                $kernel->call($session, 'spawn_http_comp', $to_spawn);
+            }
+        }
+    }
+
     # Find the http component with the least load. 
     my $min = undef;
     my $fetcher_id;
 
-    while (my($id, $data) = each %{$obj->fetchers}) {
+    while (my($id, $data) = each %fetchers) {
         my $size = scalar keys %{$data->{jobs}};
+
+        if ($data->{shutdown}) {
+            delete $fetchers{$id} if $size == 0;
+            next;
+        }
         if (!defined $min || $min > $size) {
             $fetcher_id = $id;
             $min = $size;
@@ -362,39 +449,46 @@ sub dispatch_http_fetch
 
     # Resolve this URI first (check flags for sanity. we don't want to
     # get into an infinite loop or something)
-    if ($uri->host =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {
-        $job->notes(host_name => $uri->host);
-        $job->notes(host_ip   => $uri->host);
-    } elsif (! $job->notes('dns_resolved') &&
-        $uri->host !~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {
-        my $host     = $uri->host;
-        my $dnscache = $obj->dns_cache();
-        my $host_ip  = $dnscache->get($host);
-
-        if ($host_ip) {
-            $job->notes(dns_resolved => 1);
-        } else {
-            $kernel->call($session, 'register_dns_request', $host, $job);
-            return $kernel->post(
-                $obj->dns_comp_alias,
-                'resolve',
-                'handle_dns_response',
-                $host
-            );
-        }
-
-        # We cache failed responses as well. Check if this is such case
-        if ($host_ip eq '0.0.0.0') {
-            my $err = "Could not connect to " .
-                "$uri (No address associated with $host (CACHED))";
-            return $kernel->yield('fake_error_response', $job, $err);
-        }
-
-        $job->notes(host_name => $host);
-        $job->notes(host_ip   => $host_ip);
-    }
 
     my $host_ip = $job->notes('host_ip');
+    
+    if (!$host_ip) {
+        if ($uri->host =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {
+            $host_ip = $uri->host;
+            $job->notes(host_name => $uri->host);
+            $job->notes(host_ip   => $uri->host);
+        } elsif (! $job->notes('dns_resolved') &&
+            $uri->host !~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {
+            my $host     = $uri->host;
+            my $dnscache = $obj->dns_cache();
+            $host_ip  = $dnscache ? $dnscache->get($host) : undef;
+            if ($host eq 'localhost') {
+                $host_ip = '127.0.0.1';
+                $dnscache->set($host, $host_ip) if $dnscache;
+            }
+    
+            if ($host_ip) {
+                $job->notes(dns_resolved => 1);
+                $job->notes(host_name => $host);
+                $job->notes(host_ip   => $host_ip);
+            } else {
+                $kernel->call($session, 'register_dns_request', $host, $job);
+                return $kernel->post(
+                    $obj->dns_comp_alias,
+                    'resolve',
+                    'handle_dns_response',
+                    $host
+                );
+            }
+
+            # We cache failed responses as well. Check if this is such case
+            if ($host_ip eq '0.0.0.0') {
+                my $err = "Could not connect to " .
+                    "$uri (No address associated with $host (CACHED))";
+                return $kernel->yield('fake_error_response', $job, $err);
+            }
+        }
+    }
 
     # At this point, either the uri that was requested was already an IP
     # address, or we have successfully resolved the hostname.
@@ -451,6 +545,14 @@ sub handle_http_response
     }
     $response->request($request);
 
+    # Check for possibile case of being throttled. 400 or 408
+    if ($obj->auto_throttle) {
+        # if we're auto throttling, and we've reduced the number of
+        # components thus far, c
+
+        $obj->throttle_data->add($response->code =~ /^40(?:0|8)/ ? 1 : 0);
+    }
+
     $kernel->post($obj->handler_alias, 'handle_response', $job) or
         die "Failed to post event 'handle_response' to '" . $obj->handler_alias . "'";
     $kernel->yield('finalize_job', $job);
@@ -474,7 +576,7 @@ EOHTML
     my $request = HTTP::Request->new(GET => $job->uri);
     $request->content('DUMMY-REQUEST');
     $response->request($request);
-    $kernel->yield('handle_dns_response', [ $request, $job ], [ $response ]);
+    $kernel->yield('handle_http_response', [ $request, $job ], [ $response ]);
 }
 
 sub DESTROY {}
@@ -546,8 +648,11 @@ sub handle_dns_response
 
         if ($ip) {
             Xango::debug("[handle_dns_response]: $request_address resolved to $ip.");
-            $obj->dns_cache->set($request_address, $ip);
+            $obj->dns_cache->set($request_address, $ip)
+                if $obj->dns_cache;
             foreach my $job (@$requests) {
+                $job->notes(host_name => $request_address);
+                $job->notes(host_ip   => $ip);
                 $kernel->yield('dispatch_http_fetch', $job);
             }
             return 1; # yay
@@ -581,8 +686,7 @@ sub handle_dns_response
             );
         }
     }
-    my $dnscache = $obj->dns_cache;
-    $dnscache->set($request_address, '0.0.0.0');
+    $obj->dns_cache->set($request_address, '0.0.0.0') if $obj->dns_cache;
     return undef;
 }
 
@@ -633,6 +737,12 @@ The POE session alias for the broker.
 Path to the config file used by Xango::Config. Values in this file will be 
 used as the default value for other variables. 
 
+=item EnableDnsCache
+
+Flag to tell Xango if DNS caching should be used. By default it's on.
+Turn this off for short-running Xango process that want to avoid
+repeated DNS failures caused by a cached DNS query failure.
+
 =item DnsCacheClass (scalar)
 
 Perl class to use for DNS cache.
@@ -746,6 +856,11 @@ Spawns the component to resolve DNS lookups.
 
 Send a signal to the specified HTTP component.
 
+=head2 shutdown_broker
+
+Sets the shutdown flag on the broker. The broker will attempt to shutdown
+gracefully.
+
 =head2 handle_dns_response
 
 =head2 handle_http_response
@@ -764,11 +879,12 @@ Handle a given signal. The signal name is in ARG0.
 
 =head1 SEE ALSO
 
-L<Xango::Broker::Pull> L<Xango::Broker::Push>
+L<Xango::Broker::Pull|Xango::Broker::Pull> L<Xango::Broker::Push|Xango::Broker::Push>
 
 =head1 AUTHOR
 
-Copyright (c) 2005 Daisuke Maki E<lt>dmaki@cpan.orgE<gt>. All rights reserved.
+Copyright (c) 2005-2006 Daisuke Maki E<lt>dmaki@cpan.orgE<gt>.
+All rights reserved.
 Development funded by Brazil, Ltd. E<lt>http://b.razil.jpE<gt>
 
 =cut 
